/* /////////////////////////////////////////////////////////////////////////////////////////////////
Life#IdlePoser v6 by MacGruber.
State-based idle animation system with anchoring.
https://www.patreon.com/MacGruber_Laboratory

Licensed under CC BY-SA after EarlyAccess ended. (see https://creativecommons.org/licenses/by-sa/4.0/)

Version 6 - 2020-12-05
	Implemented triggers at beging and end of the transtion, both for entering and exiting a state.
	Implemented external control triggers: SwitchState, SetStateMask and TriggerSync.
	Added option for InfiniteDuration for states.
	Updated MacGruber.Utils and UI tech overhaul.

Version 5 - 2020-08-25
	Fixing morph captures for VaM 1.20.
	Requiring VaM 1.20 now.

Version 4 - 2020-07-19
	Improve curve debug render, now with colors! Also added a color legend.
	Improved Bezier evaluation (performance).
	UI improvement: Menu tab bar replaces the dropdown.
	UI improvement: InputField with Label
	Implemented animation pause.
	Implemented option to freeze captures.
	Various bugfixes/polish.

Version 3 - 2020-06-21
	Implemented curve debug rendering
	Implemented Morph Capture
	UI improvement: Label with X button.
	Implemented state groups.
	Implemented intermediate points to connect paths.
	Enforce two-way transitions
	Implemented load/save as part of the scene.
	Lots of bugfixes/cleanup.

Version 2 - 2020-05-25
	Hotfix file reference issue.

Version 1 - 2020-05-24
	Initial prototype release.

///////////////////////////////////////////////////////////////////////////////////////////////// */

#if !VAM_GT_1_20
	#error IdlePoser requires VaM 1.20 or newer!
#endif

using UnityEngine;
using UnityEngine.Events;
using System;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
	public partial class IdlePoser : MVRScript
	{
		private const int MAX_STATES = 4;
		private static readonly int[] DISTANCE_SAMPLES = new int[] { 0, 0, 0, 11, 20};

		private const int STATETYPE_REGULARSTATE = 0;
		private const int STATETYPE_CONTROLPOINT = 1;
		private const int STATETYPE_INTERMEDIATE = 2;
		private const int NUM_STATETYPES = 3;

		private Dictionary<string, State> myStates = new Dictionary<string, State>();
		private List<ControlCapture> myControlCaptures = new List<ControlCapture>();
		private List<MorphCapture> myMorphCaptures = new List<MorphCapture>();
		private List<State> myTransition = new List<State>(8);
		private List<State> myCurrentTransition = new List<State>(MAX_STATES);
		private List<TriggerActionDiscrete> myTriggerActionsNeedingUpdate = new List<TriggerActionDiscrete>();
		private State myCurrentState;
		private State myNextState;
		private State myBlendState = State.CreateBlendState();

		private float myDuration = 1.0f;
		private float myClock = 0.0f;
		private uint myStateMask = 0;
		private bool myPlayMode = false;
		private bool myPaused = false;
		private bool myNeedRefresh = false;
		private bool myWasLoading = true;

		private JSONStorableString mySwitchState;
		private JSONStorableString mySetStateMask;

		public override void Init()
		{
			myWasLoading = true;
			myClock = 0.0f;

			ControlCapture lHand = new ControlCapture(this, "lHandControl");
			if (lHand.IsValid())
				myControlCaptures.Add(lHand);
			ControlCapture rHand = new ControlCapture(this, "rHandControl");
			if (rHand.IsValid())
				myControlCaptures.Add(rHand);

			InitUI();


			mySwitchState = new JSONStorableString("SwitchState", "", SwitchStateAction);
			mySwitchState.isStorable = mySwitchState.isRestorable = false;
			RegisterString(mySwitchState);

			mySetStateMask = new JSONStorableString("SetStateMask", "", SetStateMaskAction);
			mySetStateMask.isStorable = mySetStateMask.isRestorable = false;
			RegisterString(mySetStateMask);

			Utils.SetupAction(this, "TriggerSync", TriggerSyncAction);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			OnDestroyUI();

			foreach (var s in myStates)
			{
				State state = s.Value;
				state.EnterBeginTrigger.Remove();
				state.EnterEndTrigger.Remove();
				state.ExitBeginTrigger.Remove();
				state.ExitEndTrigger.Remove();
			}
		}

		private State CreateState(string name)
		{
			State s = new State(this, name) {
				myWaitDurationMin = 3.0f,
				myWaitDurationMax = 5.0f,
				myTransitionDuration = 0.5f,
				myStateType = STATETYPE_REGULARSTATE
			};
			CaptureState(s);
			myStates[name] = s;
			return s;
		}

		private void SetState(State state)
		{
			myCurrentState = state;
			myNextState = null;

			myClock = 0.0f;
			if (state.myWaitInfiniteDuration)
				myDuration = float.MaxValue;
			else
				myDuration = UnityEngine.Random.Range(state.myWaitDurationMin, state.myWaitDurationMax);
			myCurrentTransition.Clear();
		}

		private void SetTransition(float duration = -1.0f)
		{
			float d = 0.0f;
			int entryCount = Mathf.Min(myTransition.Count, MAX_STATES);
			for (int i=0; i<entryCount; ++i)
			{
				d += myTransition[i].myTransitionDuration;
				if (i > 0 && !myTransition[i].IsControlPoint)
				{
					entryCount = i+1;
					break;
				}
			}

			myCurrentTransition.Clear();
			for (int i=0; i<entryCount; ++i)
				myCurrentTransition.Add(myTransition[i]);

			myClock = 0.0f;
			myDuration = (duration < 0) ? d : duration;
			myDuration = Mathf.Max(myDuration, 0.001f);
			myNextState = myTransition[entryCount-1];
			for (int i=0; i<myControlCaptures.Count; ++i)
				myControlCaptures[i].SetTransition(myTransition, entryCount);
			for (int i=0; i<myMorphCaptures.Count; ++i)
				myMorphCaptures[i].SetTransition(myTransition, entryCount);

			if (myTransition.Count == entryCount)
				myTransition.Clear();
			else
				myTransition.RemoveRange(0, entryCount-1);

			if (myCurrentState.ExitBeginTrigger != null)
				myCurrentState.ExitBeginTrigger.Trigger(myTriggerActionsNeedingUpdate);
			myNextState.EnterBeginTrigger.Trigger(myTriggerActionsNeedingUpdate);
		}

		private void SetRandomTransition()
		{
			List<State> states = new List<State>(16);
			myTransition.Clear();
			myTransition.Add(myCurrentState);
			myNextState = null;
			GatherStates(1, states);

			int i;
			float sum = 0.0f;
			for (i=0; i<states.Count; ++i)
				sum += states[i].myProbability;
			if (sum == 0.0f)
				return;

			float threshold = UnityEngine.Random.Range(0.0f, sum);
			sum = 0.0f;
			for (i=0; i<states.Count-1; ++i)
			{
				sum += states[i].myProbability;
				if (threshold <= sum)
					break;
			}
			GatherTransition(1, i, 0);
			SetTransition();
		}

		private void SetBlendTransition(State state)
		{
			myTransition.Clear();
			myTransition.Add(myCurrentState);
			if (myCurrentState != null)
			{
				List<State> states = new List<State>(16);
				myNextState = state;
				GatherStates(1, states);
				List<int> indices = new List<int>(4);
				for (int i=0; i<states.Count; ++i)
				{
					if (states[i] == state)
						indices.Add(i);
				}
				if (indices.Count > 0)
				{
					int selected = UnityEngine.Random.Range(0, indices.Count);
					GatherTransition(1, indices[selected], 0);
				}
			}

			CaptureState(myBlendState);
			myTransition[0] = myBlendState;
			if (myTransition.Count == 1) // Did not find transition....fake one
				myTransition.Add(state);
			if (myCurrentState != null && myTransition.Count > 1)
				myBlendState.myTransitionDuration = myCurrentState.myTransitionDuration;
			else
				myBlendState.myTransitionDuration = 0.2f;

			myBlendState.AssignOutTriggers(myCurrentState);
			SetState(myBlendState);
			SetTransition();
		}

		private void GatherStates(int transitionLength, List<State> states)
		{
			State source = myTransition[0];
			State current = myTransition[myTransition.Count-1];
			for (int i=0; i<current.myTransitions.Count; ++i)
			{
				State next = current.myTransitions[i];
				if (myTransition.Contains(next))
					continue;

				if (next.IsRegularState || next == myNextState)
				{
					if (next.myProbability < 0.01f && next != myNextState)
						continue;
					if (myStateMask != 0 && (myStateMask & next.StateMask) == 0 && next != myNextState)
						continue;
					if (source.myStateGroup == 0 || source.myStateGroup != next.myStateGroup || next == myNextState)
					{
						states.Add(next);
					}
				}
				else if (next.IsControlPoint)
				{
					if (transitionLength >= MAX_STATES-1)
						continue;
					myTransition.Add(next);
					GatherStates(transitionLength+1, states);
					myTransition.RemoveAt(myTransition.Count-1);
				}
				else // next.IsIntermediate
				{
					myTransition.Add(next);
					GatherStates(1, states);
					myTransition.RemoveAt(myTransition.Count-1);
				}
			}
		}

		private int GatherTransition(int transitionLength, int selected, int index)
		{
			State source = myTransition[0];
			State current = myTransition[myTransition.Count-1];
			for (int i=0; i<current.myTransitions.Count; ++i)
			{
				State next = current.myTransitions[i];
				if (myTransition.Contains(next))
					continue;

				if (next.IsRegularState || next == myNextState)
				{
					if (next.myProbability < 0.01f && next != myNextState)
					{
						continue;
					}
					else if (myStateMask != 0 && (myStateMask & next.StateMask) == 0 && next != myNextState)
					{
						continue;
					}
					else if (source.myStateGroup != 0 && source.myStateGroup == next.myStateGroup && next != myNextState)
					{
						continue;
					}
					else if (index == selected)
					{
						myTransition.Add(next);
						return -1;
					}
					else
					{
						++index;
					}
				}
				else if (next.IsControlPoint)
				{
					if (transitionLength >= MAX_STATES-1)
						continue;
					myTransition.Add(next);
					index = GatherTransition(transitionLength+1, selected, index);
					if (index == -1)
						return -1;
					myTransition.RemoveAt(myTransition.Count-1);
				}
				else // next.IsIntermediate
				{
					myTransition.Add(next);
					index = GatherTransition(1, selected, index);
					if (index == -1)
						return -1;
					myTransition.RemoveAt(myTransition.Count-1);
				}
			}
			return index;
		}

		private void BlendToRandomState(float duration)
		{
			List<State> possible = new List<State>(myStates.Count);
			foreach (var state in myStates)
			{
				if (state.Value.IsRegularState)
					possible.Add(state.Value);
			}
			int idx = UnityEngine.Random.Range(0, possible.Count);

			CaptureState(myBlendState);
			myTransition.Clear();
			myTransition.Add(myBlendState);
			myTransition.Add(possible[idx]);

			myBlendState.AssignOutTriggers(myCurrentState);
			SetState(myBlendState);
			SetTransition(duration);
		}

		private void CaptureState(State state)
		{
			for (int i=0; i<myControlCaptures.Count; ++i)
				myControlCaptures[i].CaptureEntry(state);
			for (int i=0; i<myMorphCaptures.Count; ++i)
				myMorphCaptures[i].CaptureEntry(state);
		}

		private void SwitchStateAction(string v)
		{
			mySwitchState.valNoCallback = string.Empty;

			State state;
			if (myStates.TryGetValue(v, out state))
				SetBlendTransition(state);
			else
				SuperController.LogError("IdlePoser: Can't switch to unknown state '"+v+"'!");
		}

		private void SetStateMaskAction(string v)
		{
			mySetStateMask.valNoCallback = string.Empty;

			myStateMask = 0;
			bool invert = false;
			bool error = false;
			for (int i=0; i<v.Length; ++i)
			{
				char c = v[i];
				if (i == 0 && c == '!')
					invert = true;
				else if (c >= 'A' && c <= 'H')
					myStateMask |= 1u << (c - 'A' + 1);
				else if (c >= 'a' && c <= 'h')
					myStateMask |= 1u << (c - 'a' + 1);
				else if (c == 'N' || c == 'n')
					myStateMask |= 1u << 0;
				else
					error = true;
			}

			if (error)
			{
				SuperController.LogError("IdlePoser: SetStateMask set to invalid data: '"+v+"'");
				return;
			}

			if (myStateMask == 0)
				return;
			if (invert)
				myStateMask = ~myStateMask;

			if (myTransition.Count > 0 && (myTransition[myTransition.Count-1].StateMask & myStateMask) == 0)
				SetRandomTransition();
			else if (myNextState != null && (myNextState.StateMask & myStateMask) == 0)
				SetRandomTransition();
			else if (myCurrentState != null && (myCurrentState.StateMask & myStateMask) == 0)
				SetRandomTransition();
		}

		private void TriggerSyncAction()
		{
			if (myCurrentState != null && myNextState == null
			 && myClock >= myDuration && !myCurrentState.myWaitInfiniteDuration
			 && myCurrentState.myWaitForSync && !myPaused)
			{
				if (myTransition.Count == 0)
					SetRandomTransition();
				else
					SetTransition();
			}
		}

		private void Update()
		{
			bool isLoading = SuperController.singleton.isLoading;
			bool isOrWasLoading = isLoading || myWasLoading;
			myWasLoading = isLoading;
			if (isOrWasLoading)
				return;

			if (myNeedRefresh)
			{
				UIRefreshMenu();
				myNeedRefresh = false;
			}
			DebugUpdateUI();

			for (int i=0; i<myTriggerActionsNeedingUpdate.Count; ++i)
				myTriggerActionsNeedingUpdate[i].Update();
			myTriggerActionsNeedingUpdate.RemoveAll(a => !a.timerActive);

			if (myCurrentState == null)
				return;

			bool paused = myPaused && myNextState == null && myTransition.Count == 0;
			if (!paused)
				myClock = Mathf.Min(myClock + Time.deltaTime, 100000.0f);

			if (myNextState != null)
			{
				float t = Mathf.InverseLerp(0.0f, myDuration, myClock);
				t = Mathf.Clamp01(t);
				//t = SmootherStep(t);
				t = Mathf.SmoothStep(0.0f, 1.0f, t);
				for (int i=0; i<myControlCaptures.Count; ++i)
					myControlCaptures[i].UpdateTransition(t);
				for (int i=0; i<myMorphCaptures.Count; ++i)
					myMorphCaptures[i].UpdateTransition(t);
			}
			else if (!paused || myPlayMode)
			{
				for (int i=0; i<myControlCaptures.Count; ++i)
					myControlCaptures[i].UpdateState(myCurrentState);
			}

			if (myClock >= myDuration)
			{
				if (myNextState != null)
				{
					if (myCurrentState.ExitEndTrigger != null)
						myCurrentState.ExitEndTrigger.Trigger(myTriggerActionsNeedingUpdate);
					myNextState.EnterEndTrigger.Trigger(myTriggerActionsNeedingUpdate);

					SetState(myNextState);
					if (myTransition.Count == 0)
						myMainState.valNoCallback = myCurrentState.myName;
				}
				else if (!paused && !myCurrentState.myWaitForSync)
				{
					if (myTransition.Count == 0)
						SetRandomTransition();
					else
						SetTransition();
				}
			}
		}

		private static float SmootherStep(float t)
		{
			return t * t * t * (t * (t * 6.0f - 15.0f) + 10.0f);
		}

		private void OnAtomRename(string oldid, string newid)
		{
			foreach (var s in myStates)
			{
				State state = s.Value;
				state.EnterBeginTrigger.SyncAtomNames();
				state.EnterEndTrigger.SyncAtomNames();
				state.ExitBeginTrigger.SyncAtomNames();
				state.ExitEndTrigger.SyncAtomNames();
			}
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if ((includePhysical && includeAppearance) || forceStore) // StoreType.Full
			{
				jc["idlepose"] = SavePose();
				needsStore = true;
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (restorePhysical && restoreAppearance) // StoreType.Full
			{
				if (jc.HasKey("idlepose"))
					LoadPose(jc["idlepose"].AsObject);
				myNeedRefresh = true;
			}
		}

		private JSONClass SavePose()
		{
			JSONClass jc = new JSONClass();

			// save info
			JSONClass info = new JSONClass();
			info["Format"] = "MacGruber.Life.IdlePoser";
			info["Version"].AsInt = 4;
			string creatorName = UserPreferences.singleton.creatorName;
			if (string.IsNullOrEmpty(creatorName))
				creatorName = "Unknown";
			info["Author"] = creatorName;
			jc["Info"] = info;

			// save settings
			if (myCurrentState != null)
				jc["InitialState"] = myCurrentState.myName;
			jc["Paused"].AsBool = myPlayPaused.val;
			jc["StateMask"].AsInt = (int)myStateMask;

			// save captures
			if (myControlCaptures.Count > 0)
			{
				JSONArray cclist = new JSONArray();
				for (int i=0; i<myControlCaptures.Count; ++i)
				{
					ControlCapture cc = myControlCaptures[i];
					JSONClass ccclass = new JSONClass();
					ccclass["Name"] = cc.myName;
					ccclass["ApplyPos"].AsBool = cc.myApplyPosition;
					ccclass["ApplyRot"].AsBool = cc.myApplyRotation;
					cclist.Add("", ccclass);
				}
				jc["ControlCaptures"] = cclist;
			}
			if (myMorphCaptures.Count > 0)
			{
				JSONArray mclist = new JSONArray();
				for (int i=0; i<myMorphCaptures.Count; ++i)
				{
					MorphCapture mc = myMorphCaptures[i];
					JSONClass mcclass = new JSONClass();
					mcclass["Name"] = mc.myName;
					mcclass["Apply"].AsBool = mc.myApply;
					mclist.Add("", mcclass);
				}
				jc["MorphCaptures"] = mclist;
			}

			// save states
			JSONArray slist = new JSONArray();
			foreach (var s in myStates)
			{
				State state = s.Value;
				JSONClass st = new JSONClass();
				st["Name"] = state.myName;
				st["WaitInfiniteDuration"].AsBool = state.myWaitInfiniteDuration;
				st["WaitForSync"].AsBool = state.myWaitForSync;
				st["WaitDurationMin"].AsFloat = state.myWaitDurationMin;
				st["WaitDurationMax"].AsFloat = state.myWaitDurationMax;
				st["TransitionDuration"].AsFloat = state.myTransitionDuration;
				st["Probability"].AsFloat = state.myProbability;
				st["StateType"].AsInt = state.myStateType;
				st["StateGroup"].AsInt = state.myStateGroup;

				JSONArray tlist = new JSONArray();
				for (int i=0; i<state.myTransitions.Count; ++i)
					tlist.Add("", state.myTransitions[i].myName);
				st["Transitions"] = tlist;

				if (state.myControlEntries.Count > 0)
				{
					JSONClass celist = new JSONClass();
					foreach (var e in state.myControlEntries)
					{
						ControlEntryAnchored ce = e.Value;
						JSONClass ceclass = new JSONClass();
						ceclass["PX"].AsFloat = ce.myAnchorOffset.myPosition.x;
						ceclass["PY"].AsFloat = ce.myAnchorOffset.myPosition.y;
						ceclass["PZ"].AsFloat = ce.myAnchorOffset.myPosition.z;
						Vector3 rotation = ce.myAnchorOffset.myRotation.eulerAngles;
						ceclass["RX"].AsFloat = rotation.x;
						ceclass["RY"].AsFloat = rotation.y;
						ceclass["RZ"].AsFloat = rotation.z;
						ceclass["AnchorMode"].AsInt = ce.myAnchorMode;
						if (ce.myAnchorMode >= ControlEntryAnchored.ANCHORMODE_SINGLE)
						{
							ceclass["DampingTime"].AsFloat = ce.myDampingTime;
							ceclass["AnchorAAtom"] = ce.myAnchorAAtom;
							ceclass["AnchorAControl"] = ce.myAnchorAControl;
						}
						if (ce.myAnchorMode == ControlEntryAnchored.ANCHORMODE_BLEND)
						{
							ceclass["AnchorBAtom"] = ce.myAnchorBAtom;
							ceclass["AnchorBControl"] = ce.myAnchorBControl;
							ceclass["BlendRatio"].AsFloat = ce.myBlendRatio;
						}
						celist[e.Key.myName] = ceclass;
					}
					st["ControlEntries"] = celist;
				}

				if (state.myMorphEntries.Count > 0)
				{
					JSONClass melist = new JSONClass();
					foreach (var e in state.myMorphEntries)
					{
						melist[e.Key.myName].AsFloat = e.Value;
					}
					st["MorphEntries"] = melist;
				}

				st[state.EnterBeginTrigger.Name] = state.EnterBeginTrigger.GetJSON(base.subScenePrefix);
				st[state.EnterEndTrigger.Name] = state.EnterEndTrigger.GetJSON(base.subScenePrefix);
				st[state.ExitBeginTrigger.Name] = state.ExitBeginTrigger.GetJSON(base.subScenePrefix);
				st[state.ExitEndTrigger.Name] = state.ExitEndTrigger.GetJSON(base.subScenePrefix);

				slist.Add("", st);
			}
			jc["States"] = slist;

			return jc;
		}

		private void LoadPose(JSONClass jc)
		{
			// reset
			foreach (var s in myStates)
			{
				State state = s.Value;
				state.EnterBeginTrigger.Remove();
				state.EnterEndTrigger.Remove();
				state.ExitBeginTrigger.Remove();
				state.ExitEndTrigger.Remove();
			}

			myControlCaptures.Clear();
			myMorphCaptures.Clear();
			myStates.Clear();
			myCurrentState = null;
			myNextState = null;
			myBlendState.myControlEntries.Clear();
			myBlendState.myMorphEntries.Clear();
			myClock = 0.0f;

			// load info
			int version = jc["Info"].AsObject["Version"].AsInt;

			// load captures
			if (jc.HasKey("ControlCaptures"))
			{
				JSONArray cclist = jc["ControlCaptures"].AsArray;
				for (int i=0; i<cclist.Count; ++i)
				{
					ControlCapture cc;
					if (version <= 2)
					{
						// handling legacy
						cc = new ControlCapture(this, cclist[i].Value);
					}
					else
					{
						JSONClass ccclass = cclist[i].AsObject;
						cc = new ControlCapture(this, ccclass["Name"]);
						cc.myApplyPosition = ccclass["ApplyPos"].AsBool;
						cc.myApplyRotation = ccclass["ApplyRot"].AsBool;
					}

					if (cc.IsValid())
						myControlCaptures.Add(cc);
				}
			}
			if (jc.HasKey("MorphCaptures"))
			{
				JSONArray mclist = jc["MorphCaptures"].AsArray;
				DAZCharacterSelector geometry = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;
				for (int i=0; i<mclist.Count; ++i)
				{
					MorphCapture mc;
					if (version <= 2)
					{
						// handling legacy
						mc = new MorphCapture(geometry, mclist[i].Value);
					}
					else
					{
						JSONClass mcclass = mclist[i].AsObject;
						mc = new MorphCapture(geometry, mcclass["Name"]);
						mc.myApply = mcclass["Apply"].AsBool;
					}

					if (mc.IsValid())
						myMorphCaptures.Add(mc);
				}
			}

			// load states
			JSONArray slist = jc["States"].AsArray;
			for (int i=0; i<slist.Count; ++i)
			{
				// load state
				JSONClass st = slist[i].AsObject;
				State state;
				if (version == 1)
				{
					// handling legacy
					state = new State(this, st["Name"]) {
						myWaitDurationMin = st["WaitDurationMin"].AsFloat,
						myWaitDurationMax = st["WaitDurationMax"].AsFloat,
						myTransitionDuration = st["TransitionDuration"].AsFloat,
						myProbability = 0.5f,
						myStateType = st["IsTransition"].AsBool ? STATETYPE_CONTROLPOINT : STATETYPE_REGULARSTATE,
						myStateGroup = 0
					};
				}
				else
				{
					state = new State(this, st["Name"]) {
						myWaitInfiniteDuration = st["WaitInfiniteDuration"].AsBool,
						myWaitForSync = st["WaitForSync"].AsBool,
						myWaitDurationMin = st["WaitDurationMin"].AsFloat,
						myWaitDurationMax = st["WaitDurationMax"].AsFloat,
						myTransitionDuration = st["TransitionDuration"].AsFloat,
						myProbability = st["Probability"].AsFloat,
						myStateType = st["StateType"].AsInt,
						myStateGroup = st["StateGroup"].AsInt
					};
				}

				state.EnterBeginTrigger.RestoreFromJSON(st, base.subScenePrefix, base.mergeRestore, true);
				state.EnterEndTrigger.RestoreFromJSON(st, base.subScenePrefix, base.mergeRestore, true);
				state.ExitBeginTrigger.RestoreFromJSON(st, base.subScenePrefix, base.mergeRestore, true);
				state.ExitEndTrigger.RestoreFromJSON(st, base.subScenePrefix, base.mergeRestore, true);


				if (myStates.ContainsKey(state.myName))
					continue;
				myStates[state.myName] = state;

				// load control captures
				if (myControlCaptures.Count > 0)
				{
					JSONClass celist = st["ControlEntries"].AsObject;
					foreach (string ccname in celist.Keys)
					{
						ControlCapture cc = myControlCaptures.Find(x => x.myName == ccname);
						if (cc == null)
							continue;

						JSONClass ceclass = celist[ccname].AsObject;
						ControlEntryAnchored ce = new ControlEntryAnchored();
						ce.myAnchorOffset.myPosition.x = ceclass["PX"].AsFloat;
						ce.myAnchorOffset.myPosition.y = ceclass["PY"].AsFloat;
						ce.myAnchorOffset.myPosition.z = ceclass["PZ"].AsFloat;
						Vector3 rotation;
						rotation.x = ceclass["RX"].AsFloat;
						rotation.y = ceclass["RY"].AsFloat;
						rotation.z = ceclass["RZ"].AsFloat;
						ce.myAnchorOffset.myRotation.eulerAngles = rotation;
						ce.myAnchorMode = ceclass["AnchorMode"].AsInt;
						if (ce.myAnchorMode >= ControlEntryAnchored.ANCHORMODE_SINGLE)
						{
							ce.myDampingTime = ceclass["DampingTime"].AsFloat;
							ce.myAnchorAAtom = ceclass["AnchorAAtom"].Value;
							ce.myAnchorAControl = ceclass["AnchorAControl"].Value;
						}
						if (ce.myAnchorMode == ControlEntryAnchored.ANCHORMODE_BLEND)
						{
							ce.myAnchorBAtom = ceclass["AnchorBAtom"].Value;
							ce.myAnchorBControl = ceclass["AnchorBControl"].Value;
							ce.myBlendRatio = ceclass["BlendRatio"].AsFloat;
						}
						ce.Initialize(this);

						state.myControlEntries.Add(cc, ce);
					}
					for (int j=0; j<myControlCaptures.Count; ++j)
					{
						if (!state.myControlEntries.ContainsKey(myControlCaptures[j]))
							myControlCaptures[j].CaptureEntry(state);
					}
				}

				// load morph captures
				if (myMorphCaptures.Count > 0)
				{
					JSONClass melist = st["MorphEntries"].AsObject;
					foreach (string mcname in melist.Keys)
					{
						MorphCapture mc = myMorphCaptures.Find(x => x.myName == mcname);
						if (mc == null)
							continue;
						float me = melist[mcname].AsFloat;
						state.myMorphEntries.Add(mc, me);
					}
					for (int j=0; j<myMorphCaptures.Count; ++j)
					{
						if (!state.myMorphEntries.ContainsKey(myMorphCaptures[j]))
							myMorphCaptures[j].CaptureEntry(state);
					}
				}
			}

			// load transitions
			for (int i=0; i<slist.Count; ++i)
			{
				JSONClass st = slist[i].AsObject;
				State source;
				if (!myStates.TryGetValue(st["Name"], out source))
					continue;

				JSONArray tlist = st["Transitions"].AsArray;
				for (int j=0; j<tlist.Count; ++j)
				{
					State target;
					if (myStates.TryGetValue(tlist[j].Value, out target))
						source.myTransitions.Add(target);
				}
			}

			// sanity check
			foreach (var s in myStates)
			{
				State source = s.Value;
				for (int i=0; i<source.myTransitions.Count; ++i)
				{
					State target = source.myTransitions[i];
					if (!target.myTransitions.Contains(source))
						SuperController.LogError("IdlePoser: One-way transition loaded! " + source.myName + " => " + target.myName);
				}
			}

			// load settings
			myPlayPaused.valNoCallback = jc.HasKey("Paused") && jc["Paused"].AsBool;
			myPlayPaused.setCallbackFunction(myPlayPaused.val);

			int stateMask = jc["StateMask"].AsInt;
			myStateMask = (uint)stateMask;

			// blend to initial state
			if (jc.HasKey("InitialState"))
			{
				State initial;
				if (myStates.TryGetValue(jc["InitialState"].Value, out initial))
				{
					SetState(initial);
					myMainState.valNoCallback = initial.myName;
				}
			}
		}

		// =======================================================================================


		private Atom GetSelfOrAtomById(string atomName)
		{
			return atomName == "[Self]" ? containingAtom : GetAtomById(atomName);
		}

		/*private static void DebugLogList(string message, List<string> list)
		{
			for (int i=0; i<list.Count; ++i)
				message += list[i] + ", ";
			SuperController.LogMessage(message);
		}

		private static void DebugLogList(string message, List<State> list)
		{
			for (int i=0; i<list.Count; ++i)
				message += list[i].myName + ", ";
			SuperController.LogMessage(message);
		}*/

		// =======================================================================================


		private class State
		{
			public string myName;
			public float myWaitDurationMin;
			public float myWaitDurationMax;
			public float myTransitionDuration;
			public float myProbability = 0.5f;
			public int myStateType;
			public int myStateGroup;
			public bool myWaitInfiniteDuration = false;
			public bool myWaitForSync = false;
			public uint myDebugIndex = 0;
			public Dictionary<ControlCapture, ControlEntryAnchored> myControlEntries = new Dictionary<ControlCapture, ControlEntryAnchored>();
			public Dictionary<MorphCapture, float> myMorphEntries = new Dictionary<MorphCapture, float>();
			public List<State> myTransitions = new List<State>();
			public EventTrigger EnterBeginTrigger;
			public EventTrigger EnterEndTrigger;
			public EventTrigger ExitBeginTrigger;
			public EventTrigger ExitEndTrigger;

			public bool IsRegularState { get { return myStateType == STATETYPE_REGULARSTATE; } }
			public bool IsControlPoint { get { return myStateType == STATETYPE_CONTROLPOINT; } }
			public bool IsIntermediate { get { return myStateType == STATETYPE_INTERMEDIATE; } }
			public uint StateMask { get { return 1u << myStateGroup; } }

			private State(string name)
			{
				myName = name;
				// do NOT init event triggers
			}

			public State(MVRScript script, string name)
			{
				myName = name;
				EnterBeginTrigger = new EventTrigger(script, "OnEnterBegin", name);
				EnterEndTrigger = new EventTrigger(script, "OnEnterEnd", name);
				ExitBeginTrigger = new EventTrigger(script, "OnExitBegin", name);
				ExitEndTrigger = new EventTrigger(script, "OnExitEnd", name);
			}

			public static State CreateBlendState()
			{
				return new State("BlendState") {
					myWaitDurationMin = 0.0f,
					myWaitDurationMax = 0.0f,
					myTransitionDuration = 0.2f,
					myStateType = STATETYPE_CONTROLPOINT
				};
			}

			public void AssignOutTriggers(State other)
			{
				ExitBeginTrigger = other?.ExitBeginTrigger;
				ExitEndTrigger = other?.ExitEndTrigger;
			}
		}

		private class ControlCapture
		{
			public string myName;
			private IdlePoser myPlugin;
			private Transform myTransform;
			private ControlEntryAnchored[] myTransition = new ControlEntryAnchored[MAX_STATES];
			private int myEntryCount = 0;
			public bool myApplyPosition = true;
			public bool myApplyRotation = true;

			private static Quaternion[] ourTempQuaternions = new Quaternion[MAX_STATES-1];
			private static float[] ourTempDistances = new float[DISTANCE_SAMPLES[DISTANCE_SAMPLES.Length-1] + 2];

			public ControlCapture(IdlePoser plugin, string control)
			{
				myPlugin = plugin;
				myName = control;
				FreeControllerV3 controller = plugin.containingAtom.GetStorableByID(control) as FreeControllerV3;
				if (controller != null)
					myTransform = controller.transform;
			}

			public void CaptureEntry(State state)
			{
				ControlEntryAnchored entry;
				if (!state.myControlEntries.TryGetValue(this, out entry))
				{
					entry = new ControlEntryAnchored();
					entry.Initialize(myPlugin);
					state.myControlEntries[this] = entry;
				}
				entry.Capture(myTransform.position, myTransform.rotation);
			}

			public void SetTransition(List<State> states, int entryCount)
			{
				myEntryCount = entryCount;
				for (int i=0; i<myEntryCount; ++i)
				{
					if (!states[i].myControlEntries.TryGetValue(this, out myTransition[i]))
					{
						CaptureEntry(states[i]);
						myTransition[i] = states[i].myControlEntries[this];
					}
				}
			}

			public void UpdateTransition(float t)
			{
				for (int i=0; i<myEntryCount; ++i)
					myTransition[i].Update();

				//t = ArcLengthParametrization(t);

				if (myApplyPosition)
				{
					switch (myEntryCount)
					{
						case 4:	myTransform.position = EvalBezierCubicPosition(t);          break;
						case 3: myTransform.position = EvalBezierQuadraticPosition(t);      break;
						case 2: myTransform.position = EvalBezierLinearPosition(t);         break;
						default: myTransform.position = myTransition[0].myEntry.myPosition; break;
					}
				}
				if (myApplyRotation)
				{
					switch (myEntryCount)
					{
						case 4: myTransform.rotation = EvalBezierCubicRotation(t);          break;
						case 3: myTransform.rotation = EvalBezierQuadraticRotation(t);      break;
						case 2: myTransform.rotation = EvalBezierLinearRotation(t);         break;
						default: myTransform.rotation = myTransition[0].myEntry.myRotation; break;
					}
				}
			}

			private float ArcLengthParametrization(float t)
			{
				if (myEntryCount <= 2 || myEntryCount > 4)
					return t;

				int numSamples = DISTANCE_SAMPLES[myEntryCount];
				float numLines = (float)(numSamples+1);
				float distance = 0.0f;
				Vector3 previous = myTransition[0].myEntry.myPosition;
				ourTempDistances[0] = 0.0f;

				if (myEntryCount == 3)
				{
					for (int i=1; i<=numSamples; ++i)
					{
						Vector3 current = EvalBezierQuadraticPosition(i / numLines);
						distance += Vector3.Distance(previous, current);
						ourTempDistances[i] = distance;
						previous = current;
					}
				}
				else
				{
					for (int i=1; i<=numSamples; ++i)
					{
						Vector3 current = EvalBezierCubicPosition(i / numLines);
						distance += Vector3.Distance(previous, current);
						ourTempDistances[i] = distance;
						previous = current;
					}
				}

				distance += Vector3.Distance(previous, myTransition[myEntryCount-1].myEntry.myPosition);
				ourTempDistances[numSamples+1] = distance;

				t *= distance;

				int idx = Array.BinarySearch(ourTempDistances, 0, numSamples+2, t);
				if (idx < 0)
				{
					idx = ~idx;
					if (idx == 0)
						return 0.0f;
					else if (idx >= numSamples+2)
						return 1.0f;
					t = Mathf.InverseLerp(ourTempDistances[idx-1], ourTempDistances[idx], t);
					return Mathf.LerpUnclamped((idx-1) / numLines, idx / numLines, t);
				}
				else
				{
					return idx / numLines;
				}
			}

			private Vector3 EvalBezierLinearPosition(float t)
			{
				return Vector3.LerpUnclamped(myTransition[0].myEntry.myPosition, myTransition[1].myEntry.myPosition, t);
			}

			private Vector3 EvalBezierQuadraticPosition(float t)
			{
				// evaluating quadratic Bézier curve using Bernstein polynomials
				float s = 1.0f - t;
				return      (s*s) * myTransition[0].myEntry.myPosition
					 + (2.0f*s*t) * myTransition[1].myEntry.myPosition
					 +      (t*t) * myTransition[2].myEntry.myPosition;
			}

			private Vector3 EvalBezierCubicPosition(float t)
			{
				// evaluating cubic Bézier curve using Bernstein polynomials
				float s = 1.0f - t;
				float t2 = t*t;
				float s2 = s*s;
				return      (s*s2) * myTransition[0].myEntry.myPosition
					 + (3.0f*s2*t) * myTransition[1].myEntry.myPosition
					 + (3.0f*s*t2) * myTransition[2].myEntry.myPosition
					 +      (t*t2) * myTransition[3].myEntry.myPosition;
			}

			private Quaternion EvalBezierLinearRotation(float t)
			{
				return Quaternion.SlerpUnclamped(myTransition[0].myEntry.myRotation, myTransition[1].myEntry.myRotation, t);
			}

			private Quaternion EvalBezierQuadraticRotation(float t)
			{
				// evaluating quadratic Bézier curve using de Casteljau's algorithm
				ourTempQuaternions[0] = Quaternion.SlerpUnclamped(myTransition[0].myEntry.myRotation, myTransition[1].myEntry.myRotation, t);
				ourTempQuaternions[1] = Quaternion.SlerpUnclamped(myTransition[1].myEntry.myRotation, myTransition[2].myEntry.myRotation, t);
				return Quaternion.SlerpUnclamped(ourTempQuaternions[0], ourTempQuaternions[1], t);
			}

			private Quaternion EvalBezierCubicRotation(float t)
			{
				// evaluating cubic Bézier curve using de Casteljau's algorithm
				for (int i=0; i<3; ++i)
					ourTempQuaternions[i] = Quaternion.SlerpUnclamped(myTransition[i].myEntry.myRotation, myTransition[i+1].myEntry.myRotation, t);
				for (int i=0; i<2; ++i)
					ourTempQuaternions[i] = Quaternion.SlerpUnclamped(ourTempQuaternions[i], ourTempQuaternions[i+1], t);
				return Quaternion.SlerpUnclamped(ourTempQuaternions[0], ourTempQuaternions[1], t);
			}

			public void UpdateState(State state)
			{
				ControlEntryAnchored entry;
				if (state.myControlEntries.TryGetValue(this, out entry))
				{
					entry.Update();
					if (myApplyPosition)
						myTransform.position = entry.myEntry.myPosition;
					if (myApplyRotation)
						myTransform.rotation = entry.myEntry.myRotation;
				}
			}

			public bool IsValid()
			{
				return myTransform != null;
			}
		}

		private struct ControlEntry
		{
			public Quaternion myRotation;
			public Vector3 myPosition;
		}

		private class ControlEntryAnchored
		{
			public const int ANCHORMODE_WORLD = 0;
			public const int ANCHORMODE_SINGLE = 1;
			public const int ANCHORMODE_BLEND = 2;

			public ControlEntry myEntry;
			public ControlEntry myAnchorOffset;
			public Transform myAnchorATransform;
			public Transform myAnchorBTransform;
			public int myAnchorMode = 1;
			public float myBlendRatio = 0.5f;
			public float myDampingTime = 0.2f;

			public string myAnchorAAtom = "[Self]";
			public string myAnchorBAtom;
			public string myAnchorAControl = "control";
			public string myAnchorBControl;

			public void Initialize(IdlePoser plugin)
			{
				GetTransforms(plugin);
				UpdateInstant();
			}

			public void AdjustAnchor(IdlePoser plugin)
			{
				UpdateInstant();
				GetTransforms(plugin);
				Capture(myEntry.myPosition, myEntry.myRotation);
			}

			private void GetTransforms(IdlePoser plugin)
			{
				if (myAnchorMode == ANCHORMODE_WORLD)
				{
					myAnchorATransform = null;
					myAnchorBTransform = null;
				}
				else
				{
					myAnchorATransform = GetTransform(plugin, myAnchorAAtom, myAnchorAControl);
					if (myAnchorMode == ANCHORMODE_BLEND)
						myAnchorBTransform = GetTransform(plugin, myAnchorBAtom, myAnchorBControl);
					else
						myAnchorBTransform = null;
				}
			}

			private Transform GetTransform(IdlePoser plugin, string atomName, string controlName)
			{
				Atom atom = plugin.GetSelfOrAtomById(atomName);
				return atom?.GetStorableByID(controlName)?.transform;
			}

			public void UpdateInstant()
			{
				float dampingTime = myDampingTime;
				myDampingTime = 0.0f;
				Update();
				myDampingTime = dampingTime;
			}

			public void Update()
			{
				if (myAnchorMode == ANCHORMODE_WORLD)
				{
					myEntry = myAnchorOffset;
				}
				else
				{
					ControlEntry anchor;
					if (myAnchorMode == ANCHORMODE_SINGLE)
					{
						if (myAnchorATransform == null)
							return;
						anchor.myPosition = myAnchorATransform.position;
						anchor.myRotation = myAnchorATransform.rotation;
					}
					else
					{
						if (myAnchorATransform == null || myAnchorBTransform == null)
							return;
						anchor.myPosition = Vector3.LerpUnclamped(myAnchorATransform.position, myAnchorBTransform.position, myBlendRatio);
						anchor.myRotation = Quaternion.SlerpUnclamped(myAnchorATransform.rotation, myAnchorBTransform.rotation, myBlendRatio);
					}
					anchor.myPosition = anchor.myPosition + anchor.myRotation * myAnchorOffset.myPosition;
					anchor.myRotation = anchor.myRotation * myAnchorOffset.myRotation;

					if (myDampingTime >= 0.001f)
					{
						float t = Mathf.Clamp01(Time.deltaTime / myDampingTime);
						myEntry.myPosition = Vector3.LerpUnclamped(myEntry.myPosition, anchor.myPosition, t);
						myEntry.myRotation = Quaternion.SlerpUnclamped(myEntry.myRotation, anchor.myRotation, t);
					}
					else
					{
						myEntry = anchor;
					}
				}
			}

			public void Capture(Vector3 position, Quaternion rotation)
			{
				myEntry.myPosition = position;
				myEntry.myRotation = rotation;

				if (myAnchorMode == ANCHORMODE_WORLD)
				{
					myAnchorOffset.myPosition = position;
					myAnchorOffset.myRotation = rotation;
				}
				else
				{
					ControlEntry root;
					if (myAnchorMode == ANCHORMODE_SINGLE)
					{
						if (myAnchorATransform == null)
							return;
						root.myPosition = myAnchorATransform.position;
						root.myRotation = myAnchorATransform.rotation;
					}
					else
					{
						if (myAnchorATransform == null || myAnchorBTransform == null)
							return;
						root.myPosition = Vector3.LerpUnclamped(myAnchorATransform.position, myAnchorBTransform.position, myBlendRatio);
						root.myRotation = Quaternion.SlerpUnclamped(myAnchorATransform.rotation, myAnchorBTransform.rotation, myBlendRatio);
					}

					myAnchorOffset.myPosition = Quaternion.Inverse(root.myRotation) * (position - root.myPosition);
					myAnchorOffset.myRotation = Quaternion.Inverse(root.myRotation) * rotation;
				}
			}
		}


		private class MorphCapture
		{
			public string myName;
			public DAZMorph myMorph;
			public DAZCharacterSelector.Gender myGender;
			private float[] myTransition = new float[MAX_STATES];
			private int myEntryCount = 0;
			public bool myApply = true;

			public MorphCapture(DAZCharacterSelector.Gender gender, DAZMorph morph)
			{
				myMorph = morph;
				myGender = gender;
				myName = myGender.ToString() + "#" + myMorph.uid;
			}

			public MorphCapture(DAZCharacterSelector geometry, string qualifiedName)
			{
				bool isFemale = qualifiedName.StartsWith("Female#");
				if (!isFemale && !qualifiedName.StartsWith("Male#"))
					return;
				GenerateDAZMorphsControlUI morphsControl = isFemale ? geometry.morphsControlFemaleUI : geometry.morphsControlMaleUI;
				string morphUID = qualifiedName.Substring(isFemale ? 7 : 5);
				myMorph = morphsControl.GetMorphByUid(morphUID);
				myGender = isFemale ? DAZCharacterSelector.Gender.Female : DAZCharacterSelector.Gender.Male;

				myName = qualifiedName;
			}

			public void CaptureEntry(State state)
			{
				state.myMorphEntries[this] = myMorph.morphValue;
			}

			public void SetTransition(List<State> states, int entryCount)
			{
				myEntryCount = entryCount;
				bool identical = true;
				float morphValue = myMorph.morphValue;
				for (int i=0; i<myEntryCount; ++i)
				{
					if (!states[i].myMorphEntries.TryGetValue(this, out myTransition[i]))
					{
						CaptureEntry(states[i]);
						myTransition[i] = morphValue;
					}
					else
					{
						identical &= (myTransition[i] == morphValue);
					}
				}
				if (identical)
					myEntryCount = 0; // nothing to do, save some performance
			}

			public void UpdateTransition(float t)
			{
				if (!myApply)
					return;

				switch (myEntryCount)
				{
					case 4:
						myMorph.morphValue = EvalBezierCubic(t);
						break;
					case 3:
						myMorph.morphValue = EvalBezierQuadratic(t);
						break;
					case 2:
						myMorph.morphValue = EvalBezierLinear(t);
						break;
					default:
						myMorph.morphValue = myTransition[0];
						break;
				}
			}

			private float EvalBezierLinear(float t)
			{
				return Mathf.LerpUnclamped(myTransition[0], myTransition[1], t);
			}

			private float EvalBezierQuadratic(float t)
			{
				// evaluating using Bernstein polynomials
				float s = 1.0f - t;
				return      (s*s) * myTransition[0]
					 + (2.0f*s*t) * myTransition[1]
					 +      (t*t) * myTransition[2];
			}

			private float EvalBezierCubic(float t)
			{
				// evaluating using Bernstein polynomials
				float s = 1.0f - t;
				float t2 = t*t;
				float s2 = s*s;
				return      (s*s2) * myTransition[0]
					 + (3.0f*s2*t) * myTransition[1]
					 + (3.0f*s*t2) * myTransition[2]
					 +      (t*t2) * myTransition[3];
			}

			public bool IsValid()
			{
				return myMorph != null;
			}
		}

	}
}